package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Activity;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataRetrievalFailureException;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.support.TransactionTemplate;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.validation.ConstraintViolationException;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
class ActivityDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private TransactionTemplate transaction;

    @Autowired
    private ActivityDao activityDao;

    private Activity storedActivity;
    private Activity storedActivitySecond;
    private Activity createdActivity;

    @BeforeAll
    void setUp() {
        storedActivity = new Activity();
        storedActivity.setTime(LocalDateTime.of(2022,3,27,0,0));
        storedActivity.setName("Activity");

        storedActivitySecond = new Activity();
        storedActivitySecond.setTime(LocalDateTime.of(2022,3,27,2,0));
        storedActivitySecond.setName("Activity second");

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();

            em.persist(storedActivity);
            em.persist(storedActivitySecond);

            em.getTransaction().commit();
        } finally {
            if (em != null) {
                em.close();
            }
        }
    }

    @Test
    @Transactional
    void createActivity() {
        createdActivity = new Activity();
        createdActivity.setTime(LocalDateTime.of(2022,3,26,0,0));
        createdActivity.setName("Created activity");
        activityDao.create(createdActivity);
        assertNotNull(activityDao.findById(createdActivity.getId()));
    }

    @Test
    void createActivityWithNullName() {
        Activity nullNameActivity = new Activity();
        nullNameActivity.setName(null);
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> activityDao.create(nullNameActivity))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void createActivityWithNullTime() {
        Activity nullDateActivity = new Activity();
        nullDateActivity.setTime(null);
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> activityDao.create(nullDateActivity))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void findByID(){
        Activity activity = activityDao.findById(storedActivity.getId());
        assertEquals(storedActivity, activity);
    }

    @Test
    void findByNullID(){
        assertThrows(IllegalArgumentException.class, () -> activityDao.findById(null));
    }

    @Test
    void findByNonExistingID(){
        Activity activity = activityDao.findById("");
        assertNull(activity);
    }

    @Test
    void findByName() {
        List<Activity> activities = activityDao.findByName(storedActivity.getName());
        List<Activity> expectedActivities = new ArrayList<>(List.of(storedActivity));
        assertEquals(expectedActivities, activities);
    }

    @Test
    void findByNullName() {
        List<Activity> activities = activityDao.findByName(null);
        assertEquals(new ArrayList<>(), activities);
    }


    @Test
    @Transactional
    void updateActivityName() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setName("New Activity");
        activityDao.update(activity);
        Activity activityWithNewName = activityDao.findById(storedActivity.getId());
        assertEquals("New Activity", activityWithNewName.getName());
    }

    @Test
    @Transactional
    void updateActivityWithNullName() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setName(null);
        assertThrows(IllegalArgumentException.class, () -> activityDao.update(activity));
    }

    @Test
    @Transactional
    void updateActivityWithEmptyName() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setName("");
        activityDao.update(activity);
        Activity activityWithEmptyName = activityDao.findById(storedActivity.getId());
        assertEquals("", activityWithEmptyName.getName());
    }

    @Test
    @Transactional
    void updateActivityTime() {
        Activity activity = activityDao.findById(storedActivity.getId());
        LocalDateTime newDate = LocalDateTime.of(2022,1,1,1,1);
        activity.setTime(newDate);
        activityDao.update(activity);
        Activity activityWithNewTime = activityDao.findById(storedActivity.getId());
        assertEquals(newDate, activityWithNewTime.getTime());
    }

    @Test
    @Transactional
    void updateActivityWithNullTime() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setTime(null);
        assertThrows(IllegalArgumentException.class, () -> activityDao.update(activity));
    }


    @Test
    @Transactional
    void deleteActivity() {
        activityDao.delete(storedActivity);
        assertNull(activityDao.findById(storedActivity.getId()));
    }

    @Test
    @Transactional
    void deleteNullActivity() {
        assertThrows(IllegalArgumentException.class, () -> activityDao.delete(null));
    }

    @Test
    @Transactional
    void deleteActivityWithoutNotInDB() {
        assertThrows(DataRetrievalFailureException.class, () -> activityDao.delete(new Activity()));
    }

    @Test
    void findAllActivities(){
        List<Activity> activitiesFromDB = activityDao.findAll();
        List<Activity> expected = List.of(storedActivity, storedActivitySecond);
        assertNotNull(activitiesFromDB);
        assertEquals(expected, activitiesFromDB);
    }
}