package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Album;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.entity.Song;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.api.function.Executable;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.support.TransactionTemplate;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import javax.validation.ConstraintViolationException;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

/**
 * @author Dávid Želník - 469323
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class AlbumDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private TransactionTemplate transaction;

    @Autowired
    private AlbumDao albumDao;

    private final Album albumOne = new Album();
    private final Album albumTwo = new Album();
    private final Song song = new Song();
    private final Band band = new Band();
    private final Musician musician = new Musician();

    @BeforeAll
    void setUp(){
        song.setSongName("Song name");
        song.setDuration(Duration.ofSeconds(30));

        musician.setName("Musician");
        musician.setType(MusicianType.GUITARIST);

        band.setName("Band");
        band.setType(BandType.ROCK);
        band.setBandMembers(new HashSet<>(List.of(musician)));

        musician.setBand(band);

        albumOne.setName("First");
        albumOne.setBand(band);
        albumOne.setSongs(new HashSet<>(List.of(song)));

        albumTwo.setName("Second");
        albumTwo.setBand(band);
        albumTwo.setSongs(new HashSet<>());

        band.setBandAlbums(new HashSet<>(List.of(albumOne, albumTwo)));

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();

            em.persist(musician);
            em.persist(band);
            em.persist(song);

            em.persist(albumOne);
            em.persist(albumTwo);

            em.getTransaction().commit();
        }
        finally {
            if(em != null){
                em.close();
            }
        }
    }

    @Test
    @Transactional
    void CreateAlbum() throws Exception {
        Album newAlbum = new Album();
        newAlbum.setName("Album");
        newAlbum.setBand(band);
    }

    @Test
    void CreateNullAlbum(){
        assertThrows(Exception.class, () -> albumDao.create(null));
    }

    @Test
    void CreateAlbumWithoutName() {
        Album createdAlbum = new Album();
        createdAlbum.setName(null);
        createdAlbum.setBand(band);
        createdAlbum.setSongs(new HashSet<>(List.of(song)));
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> albumDao.create(createdAlbum))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void CreateAlbumWithNullBand() {
        Album createdAlbum = new Album();
        createdAlbum.setName("Album");
        createdAlbum.setBand(null);
        createdAlbum.setSongs(new HashSet<>(List.of(song)));
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> albumDao.create(createdAlbum))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void CreateEmptyAlbum() {
        Album album = new Album();
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> albumDao.create(album))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);

    }

    @Test
    @Transactional
    void FindAlbumById() throws Exception {
        Album first = albumDao.findById(albumOne.getId());
        Album second = albumDao.findById(albumTwo.getId());
        assertEquals(second, albumTwo);
        assertEquals(first, albumOne);
    }

    @Test
    @Transactional
    void FindAlbumByNullId(){
        assertThrows(IllegalArgumentException.class, () -> albumDao.findById(null));
    }

    @Test
    @Transactional
    void FindAlbumByNonExistingID() {
        assertThrows(Exception.class, ()-> albumDao.findById(" "));
    }

    @Test
    @Transactional
    void FindAlbumByName() throws Exception {
        List<Album> albumsDb = albumDao.findByName(albumOne.getName());
        List<Album> expectedAlbums = new ArrayList<>(Arrays.asList(albumOne));
        assertEquals(expectedAlbums, albumsDb);
    }

    @Test
    @Transactional
    void FindAlbumByNullName() throws Exception {
        List<Album> albumsDb = albumDao.findByName(null);
        assertEquals(new ArrayList<>(), albumsDb);
    }

    @Test
    @Transactional
    void FindAlbumByEmptyName() throws Exception {
        List<Album> albumsDb = albumDao.findByName("");
        assertEquals(new ArrayList<>(), albumsDb);
    }

    @Test
    @Transactional
    void FindNonExistingAlbumByName() throws Exception {
        List<Album> albumsDb = albumDao.findByName("213456");
        assertEquals(new ArrayList<>(), albumsDb);
    }

    @Test
    @Transactional
    void UpdateAlbum() throws Exception {
        Album albumFromDb = albumDao.findById(albumOne.getId());
        albumFromDb.setName("New album");
        albumDao.update(albumFromDb);
        Album newAlbum = albumDao.findById(albumOne.getId());
        assertEquals("New album", newAlbum.getName());
    }

    @Test
    @Transactional
    void UpdateAlbumWithNullName() throws Exception {
        Album albumInDb = albumDao.findById(albumOne.getId());
        albumInDb.setName(null);
        assertThrows(Exception.class, ()-> albumDao.update(albumInDb));
    }

    @Test
    @Transactional
    void UpdateAlbumWithNullBandName() throws Exception {
        Album albumInDb = albumDao.findById(albumOne.getId());
        albumInDb.setBand(null);
        assertThrows(Exception.class,()-> albumDao.update(albumInDb));
    }

    @Test
    @Transactional
    void UpdateAlbumSongs() throws Exception {
        Album albumInDb = albumDao.findById(albumOne.getId());
        albumInDb.setSongs(new HashSet<>());
        albumDao.update(albumInDb);
        Album albumWithZeroSongs = albumDao.findById(albumOne.getId());
        assertEquals(new HashSet<>(), albumWithZeroSongs.getSongs());
    }

    @Test
    @Transactional
    void DeleteAlbum() throws Exception {
        albumDao.delete(albumOne);
        assertThrows(Exception.class,()-> albumDao.findById(albumOne.getId()));
    }

    @Test
    @Transactional
    void DeleteNullAlbum() {
        assertThrows(Exception.class, () -> albumDao.delete(null));
    }

    @Test
    @Transactional
    void DeleteAlbumNotInDb() {
        assertThrows(IllegalArgumentException.class, () -> albumDao.delete(new Album()));
    }

}
