package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;

import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataRetrievalFailureException;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.support.TransactionTemplate;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.validation.ConstraintViolationException;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertNotNull;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
class BandDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private TransactionTemplate transaction;

    @Autowired
    private BandDao bandDao;

    private Band storedBand;
    private Band storedBandSecond;
    private Band createdBand;
    private Band nullNameBand;
    private Band nullTypeBand;
    private Band nullMembersBand;
    private final Set<Musician> bandMembers = new HashSet<>();


    @BeforeAll
    void setUp() {
        storedBand = new Band();
        storedBandSecond = new Band();
        Musician musicianJohn = new Musician();
        Musician musicianJane = new Musician();
        musicianJane.setName("Jane Smith");
        musicianJane.setType(MusicianType.SINGER);
        musicianJohn.setName("John Doe");
        musicianJohn.setType(MusicianType.GUITARIST);
        bandMembers.add(musicianJane);
        bandMembers.add(musicianJohn);
        storedBand.setName("Band");
        storedBand.setType(BandType.ROCK);
        storedBand.setLogo("New Logo");
        storedBand.setBandMembers(bandMembers);
        storedBandSecond.setName("Band 2");
        storedBandSecond.setType(BandType.POP);
        storedBandSecond.setBandMembers(bandMembers);
        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();

            em.persist(musicianJane);
            em.persist(musicianJohn);
            em.persist(storedBand);
            em.persist(storedBandSecond);

            em.getTransaction().commit();
        } finally {
            if (em != null) {
                em.close();
            }
        }
    }

    @Test
    void findByID(){
        Band bandFromDB = bandDao.findById(storedBand.getId());
        assertEquals(storedBand, bandFromDB);
    }

    @Test
    void findByNullID(){
        assertThrows(IllegalArgumentException.class, () -> bandDao.findById(null));
    }

    @Test
    void findByNonExistingID(){
        Band bandFromDB = bandDao.findById("");
        assertNull(bandFromDB);
    }

    @Test
    @Transactional
    void findByName() {
        List<Band> bandsFromDB = bandDao.findByName(storedBand.getName());
        List<Band> expectedBands = new ArrayList<>(List.of(storedBand));
        assertEquals(expectedBands, bandsFromDB);
    }

    @Test
    void findByNullName() {
        List<Band> bandsFromDB = bandDao.findByName(null);
        assertEquals(new ArrayList<>(), bandsFromDB);
    }

    @Test
    @Transactional
    void createBand() {
        createdBand = new Band();
        createdBand.setBandMembers(bandMembers);
        createdBand.setName("created band");
        createdBand.setType(BandType.ROCK);
        bandDao.create(createdBand);
        assertNotNull(bandDao.findById(createdBand.getId()));
    }

    @Test
    void createBandWithNullName() {
        nullNameBand = new Band();
        nullNameBand.setBandMembers(bandMembers);
        nullNameBand.setName(null);
        nullNameBand.setType(BandType.ROCK);
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> bandDao.create(nullNameBand))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void createBandWithNullType() {
        nullTypeBand = new Band();
        nullTypeBand.setBandMembers(bandMembers);
        nullTypeBand.setName("Null type band");
        nullTypeBand.setType(null);
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> bandDao.create(nullTypeBand))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void createBandWithNullMembers() {
        nullMembersBand = new Band();
        nullMembersBand.setBandMembers(null);
        nullMembersBand.setName("Null members band");
        nullMembersBand.setType(BandType.ROCK);
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> bandDao.create(nullMembersBand))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    @Transactional
    void updateBandName() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setName("New Band");
        bandDao.update(bandFromDB);
        Band bandWithNewName = bandDao.findById(storedBand.getId());
        assertEquals("New Band", bandWithNewName.getName());
    }

    @Test
    @Transactional
    void updateBandWithNullName() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setName(null);
        assertThrows(IllegalArgumentException.class, () -> bandDao.update(bandFromDB));
    }

    @Test
    @Transactional
    void updateBandWithEmptyName() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setName("");
        bandDao.update(bandFromDB);
        Band bandWithNullName = bandDao.findById(storedBand.getId());
        assertEquals("", bandWithNullName.getName());
    }

    @Test
    @Transactional
    void updateBandType() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setType(BandType.COUNTRY);
        bandDao.update(bandFromDB);
        Band bandWithNewType = bandDao.findById(storedBand.getId());
        assertEquals(BandType.COUNTRY, bandWithNewType.getType());
    }

    @Test
    @Transactional
    void updateBandWithNullType() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setType(null);
        assertThrows(IllegalArgumentException.class, () -> bandDao.update(bandFromDB));
    }

    @Test
    @Transactional
    void deleteBand() {
        bandDao.delete(storedBand);
        assertNull(bandDao.findById(storedBand.getId()));
    }

    @Test
    @Transactional
    void deleteNullBand() {
        assertThrows(IllegalArgumentException.class, () -> bandDao.delete(null));
    }

    @Test
    @Transactional
    void deleteBandNotInDB() {
        assertThrows(DataRetrievalFailureException.class, () -> bandDao.delete(new Band()));
    }

    @Test
    @Transactional
    void findAllBands(){
        List<Band> bandsFromDB = bandDao.findAll();
        assertNotNull(bandsFromDB);
        assertThat(bandsFromDB).containsExactlyInAnyOrder(storedBand, storedBandSecond);
    }
}