package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Manager;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.BandType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataRetrievalFailureException;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.support.TransactionTemplate;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.validation.ConstraintViolationException;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertNotNull;


@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class ManagerDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private TransactionTemplate transaction;


    @Autowired
    private ManagerDao managerDao;

    private final Manager manager = new Manager();
    private final Manager manager2 = new Manager();
    private final Band band = new Band();
    private final Set<Musician> bandMembers = new HashSet<>();
    private final Band band2 = new Band();


    @BeforeAll
    void setUp() {
        // create first band
        manager.setName("Jef the manager");
        band.setName("Band 1");
        band.setType(BandType.ROCK);
        band.setLogo("New Logo");
        band.setBandMembers(bandMembers);
        manager.setManagedBands(new HashSet<>(List.of(band)));

        //create second band
        manager2.setName("Logan the manager");
        band2.setName("Original Name");
        band2.setType(BandType.ROCK);
        band2.setLogo("Original Logo");
        band2.setBandMembers(bandMembers);
        manager2.setManagedBands(new HashSet<>(List.of(band2)));


        EntityManager em = null;
        try {
            //create entity manager
            em = emf.createEntityManager();
            em.getTransaction().begin();

            em.persist(manager);
            em.persist(manager2);

            //commit added data
            em.getTransaction().commit();
        } finally {
            //after commit close manager
            if (em != null) {
                em.close();
            }
        }
    }

    @Test
    void findByID() {
        Manager managerFromDB = managerDao.findById(manager.getId());
        Manager manager2FromDB = managerDao.findById(manager2.getId());
        assertEquals(manager.getName(), managerFromDB.getName());
        assertEquals(manager2.getName(), manager2FromDB.getName());
    }

    @Test
    void findByNullID(){
        assertThrows(IllegalArgumentException.class, () -> managerDao.findById(null));
    }

    @Test
    void findByNonExistingID() {
        Manager manager = managerDao.findById("-1L");
        assertNull(manager);
    }

    @Test
    void findByName() {
        List<Manager> managerFromDB = managerDao.findByName(manager.getName());
        List<Manager> expectedManagers = new ArrayList<>(List.of(manager));
        assertEquals(expectedManagers, managerFromDB);
    }

    @Test
    void findByNullName() {
        assertThrows(IllegalArgumentException.class, () -> managerDao.findByName(null));
    }

    @Test
    void findByEmptyName() {
        List<Manager> managerFromDB = managerDao.findByName("");
        assertEquals(new ArrayList<>(), managerFromDB);
    }

    @Test
    @Transactional
    void createManager() {
        Manager testManager = new Manager();
        testManager.setName("Test Band");
        Band testBand = new Band();
        testManager.setManagedBands(new HashSet<>(List.of(testBand)));
        managerDao.create(testManager);
        assertNotNull(managerDao.findById(testManager.getId()));
    }


    @Test
    void createManagerWithNullName() {
        Manager testManager = new Manager();
        Band testBand = new Band();
        testManager.setName(null);
        testManager.setManagedBands(new HashSet<>(List.of(testBand)));
        testBand.setType(BandType.METAL);
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> managerDao.create(testManager))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    @Transactional
    void createManagerWithoutBand() {
        Manager testManager = new Manager();
        testManager.setName("Name of manager");
        managerDao.create(testManager);
        assertNotNull(managerDao.findById(testManager.getId()));
    }

    @Test
    @Transactional
    void updateManager() {
        Manager managerFromDB = managerDao.findById(manager.getId());
        managerFromDB.setName("New Manager");
        managerDao.update(managerFromDB);
        Manager newManager = managerDao.findById(manager.getId());
        assertEquals("New Manager", newManager.getName());
    }

    @Test
    @Transactional
    void updateManagerWithNullName() {
        Manager managerFromDB = managerDao.findById(manager.getId());
        managerFromDB.setName(null);
        assertThrows(IllegalArgumentException.class, () -> managerDao.update(managerFromDB));
    }

    @Test
    @Transactional
    void deleteManager() {
        managerDao.delete(manager);
        assertNull(managerDao.findById(manager.getId()));
    }
    @Test
    @Transactional
    void deleteNullManager() {
        assertThrows(IllegalArgumentException.class, () -> managerDao.delete(null));
    }

    @Test
    @Transactional
    void deleteManagerWithoutID() {
        assertThrows(DataRetrievalFailureException.class, () -> managerDao.delete(new Manager()));
    }
}
