package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.MusicianType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.support.TransactionTemplate;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import javax.validation.ConstraintViolationException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class MusicianDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private TransactionTemplate transaction;

    @Autowired
    private MusicianDao musicianDao;

    private final Musician musicianStephen = new Musician();
    private final Musician musicianMick = new Musician();

    @BeforeAll
    void setUp(){
        musicianStephen.setName("Stephen");
        musicianStephen.setType(MusicianType.GUITARIST);

        musicianMick.setName("Mick");
        musicianMick.setType(MusicianType.SINGER);

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();
            em.persist(musicianStephen);
            em.persist(musicianMick);

            em.getTransaction().commit();
        }
        finally {
            if(em != null){
                em.close();
            }
        }
    }

    @Test
    void findMusicianById(){
        Musician musicianStephenFromDb = musicianDao.findById(musicianStephen.getId());
        Musician musicianMickFromDb = musicianDao.findById(musicianMick.getId());
        assertEquals(musicianStephenFromDb, musicianStephen);
        assertEquals(musicianMickFromDb, musicianMick);
    }

    @Test
    void findMusicianByNullId(){
        assertThrows(IllegalArgumentException.class, () -> musicianDao.findById(null));
    }

    @Test
    void findMusicianByNonExistingID(){
        assertNull(musicianDao.findById(""));
    }

    @Test
    void findMusicianByName() {
        List<Musician> musicianFromDb = musicianDao.findByName(musicianStephen.getName());
        List<Musician> expectedMusicians = new ArrayList<>(List.of(musicianStephen));
        assertEquals(expectedMusicians, musicianFromDb);
    }

    @Test
    void findMusicianByNullName() {
        List<Musician> musicianFromDB = musicianDao.findByName(null);
        assertEquals(new ArrayList<>(), musicianFromDB);
    }

    @Test
    void findMusicianByEmptyName() {
        List<Musician> musicianFromDB = musicianDao.findByName("");
        assertEquals(new ArrayList<>(), musicianFromDB);
    }

    @Test
    @Transactional
    void createMusician() {
        Musician musician = new Musician();
        musician.setName("Test");
        musician.setType(MusicianType.SINGER);
        musicianDao.create(musician);
        assertNotNull(musicianDao.findById(musician.getId()));
    }

    @Test
    void createMusicianWithoutName() {
        Musician musician = new Musician();
        musician.setName(null);

        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> musicianDao.create(musician))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void createMusicianWithoutType() {
        Musician musician = new Musician();
        musician.setType(null);

        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> musicianDao.create(musician))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);

    }

    @Test
    @Transactional
    void updateMusician() {
        Musician musicianFromDb = musicianDao.findById(musicianStephen.getId());
        musicianFromDb.setName("Stephanos");
        musicianFromDb.setType(MusicianType.DRUMMER);
        musicianDao.update(musicianFromDb);
        Musician newMusician = musicianDao.findById(musicianStephen.getId());
        assertEquals("Stephanos", newMusician.getName());
        assertEquals(MusicianType.DRUMMER, newMusician.getType());
    }

    @Test
    @Transactional
    void updateMusicianWithNullName() {
        Musician musicianFromDb = musicianDao.findById(musicianStephen.getId());
        musicianFromDb.setName(null);
        assertThrows(IllegalArgumentException.class, () -> musicianDao.update(musicianFromDb));
    }

    @Test
    @Transactional
    void updateMusicianWithNullType() {
        Musician musicianFromDb = musicianDao.findById(musicianStephen.getId());
        musicianFromDb.setType(null);
        assertThrows(IllegalArgumentException.class, () -> musicianDao.update(musicianFromDb));
    }

    @Test
    @Transactional
    void deleteMusician() {
        musicianDao.delete(musicianStephen);
        assertNull(musicianDao.findById(musicianStephen.getId()));
    }

    @Test
    @Transactional
    void deleteNullMusician() {
        assertThrows(IllegalArgumentException.class, () -> musicianDao.delete(null));
    }

    @Test
    @Transactional
    void deleteNonExistingMusician() {
        assertThrows(IllegalArgumentException.class, () -> musicianDao.delete(new Musician()));
    }

    @Test
    void findAllMusicians(){
        List<Musician> musiciansFromDB = musicianDao.findAll();
        List<Musician> expected = List.of(musicianStephen, musicianMick);
        assertNotNull(musiciansFromDB);
        assertEquals(expected, musiciansFromDB);
    }

    @Test
    void findAvailableMusicians(){
        List<Musician> allMusicians = musicianDao.findAll();
        List<Musician> musiciansFromDB = musicianDao.findAvailableMusicians();
        assertNotNull(musiciansFromDB);
        assertEquals(musiciansFromDB, allMusicians);
    }
}
