package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.entity.Offer;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.support.TransactionTemplate;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class OfferDaoTest {
    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private TransactionTemplate transaction;

    @Autowired
    private OfferDao offerDao;

    private final Offer offer1 = new Offer();
    private final Offer offer2 = new Offer();

    public OfferDaoTest() {
    }


    @BeforeAll
    void setUp(){
        Musician musicianMick = new Musician();
        musicianMick.setName("Mick");
        musicianMick.setType(MusicianType.SINGER);

        Band band1 = new Band();
        band1.setName("band1");
        band1.setType(BandType.ROCK);
        band1.setBandMembers(new HashSet<>(List.of(musicianMick)));

        offer1.setName("Offer1");
        offer2.setName("Offer2");
        offer1.setBand(band1);
        offer2.setBand(band1);

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();
            em.persist(musicianMick);
            em.persist(band1);
            em.persist(offer1);
            em.persist(offer2);

            em.getTransaction().commit();
        }
        finally {
            if(em != null){
                em.close();
            }
        }
    }

    @Test
    void CreateNullOffer(){
        assertThrows(Exception.class,()-> offerDao.create(null));
    }

    @Test
    void CreateOfferWithNullBand(){
        offer1.setBand(null);
        assertThrows(Exception.class,()-> offerDao.create(offer1));
    }

    @Test
    @Transactional
    void CreateSuccessfully(){
        offerDao.create(offer1);
        assertEquals(offerDao.findById(offer1.getId()), offer1);
        assertNotEquals(offerDao.findById(offer2.getId()), offer1);
    }

    @Test
    @Transactional
    void FindByNullId(){
        assertThrows(Exception.class, ()-> offerDao.findById(null));
    }

    @Test
    void FindByNameNullName(){
        List<Offer> offersFromDB = offerDao.findByName(null);
        assertEquals(new ArrayList<>(), offersFromDB);
    }

    @Test
    void FindByNameSuccessfully() {
        Assertions.assertThat(offerDao.findByName(offer1.getName())).containsExactlyInAnyOrder(offer1);
    }

    @Test
    void UpdateNullOffer(){
        assertThrows(Exception.class,()-> offerDao.update(null));
    }

    @Test
    void UpdateNullOfferName(){
        offer1.setName(null);
        assertThrows(Exception.class,()-> offerDao.update(offer1));
    }

    @Test
    void UpdateNullOfferBrand(){
        offer1.setBand(null);
        assertThrows(Exception.class,()-> offerDao.update(null));
    }

    @Test
    @Transactional
    void UpdateSuccessfully(){
        Offer offerFromDb = offerDao.findById(offer1.getId());
        offerFromDb.setName("changed");
        offerDao.update(offerFromDb);
        Offer offerWithNewName = offerDao.findById(offer1.getId());
        assertEquals("changed", offerWithNewName.getName());
    }

    @Test
    void DeleteNull(){
        assertThrows(Exception.class, ()-> offerDao.delete(null));
    }

    @Test
    @Transactional
    void DeleteSuccessfully(){
        offerDao.delete(offer1);
        assertEquals(offerDao.findById(offer1.getId()), null);
    }

    @Test
    @Transactional
    void FindAllSuccessfully(){
        assertEquals(offerDao.findAll().stream().count(), 2);
    }
}
