package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.entity.Song;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.support.TransactionTemplate;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import javax.validation.ConstraintViolationException;
import java.time.Duration;
import java.util.*;

import static org.junit.jupiter.api.Assertions.*;

/**
 * @author Dávid Želník - 469323
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class SongDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private TransactionTemplate transaction;

    @Autowired
    private SongDao songDao;

    private final Song songOne = new Song();
    private final Song songTwo = new Song();
    private final Band band = new Band();
    private final Musician musician = new Musician();

    @BeforeAll
    void setUp(){
        musician.setName("Musician");
        musician.setType(MusicianType.GUITARIST);

        songOne.setSongName("Song One");
        songOne.setDuration(Duration.ofSeconds(30));
        songTwo.setSongName("Song Two");
        songTwo.setDuration(Duration.ofSeconds(30));

        band.setName("Band");
        band.setType(BandType.ROCK);
        band.setBandMembers(new HashSet<>(List.of(musician)));

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();


            em.persist(band);
            em.persist(musician);
            em.persist(songTwo);
            em.persist(songOne);

            em.getTransaction().commit();
        }
        finally {
            if(em != null){
                em.close();
            }
        }
    }

    @Test
    @Transactional
    void CreateSong() {
        Song newSong = new Song();
        newSong.setSongName("Song");
        newSong.setDuration(Duration.ofSeconds(35));
        songDao.create(newSong);
        assertNotNull(songDao.findById(newSong.getId()));
    }

    @Test
    void CreateSongWithNullName() {
        Song createdSong = new Song();
        createdSong.setSongName(null);
        createdSong.setDuration(Duration.ofSeconds(35));
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> songDao.create(createdSong))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void CreateSongWithNullTime() {
        Song createdSong = new Song();
        createdSong.setSongName("Songs");
        createdSong.setDuration(null);
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> songDao.create(createdSong))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    void CreateEmptySong() {
        Song song = new Song();
        Assertions.assertThatThrownBy(() ->
                transaction.executeWithoutResult(transactionStatus -> songDao.create(song))
        ).hasRootCauseInstanceOf(ConstraintViolationException.class);
    }

    @Test
    @Transactional
    void FindSongById() {
        Song first = songDao.findById(songOne.getId());
        Song second = songDao.findById(songTwo.getId());
        assertEquals(second, songTwo);
        assertEquals(first, songOne);
    }

    @Test
    @Transactional
    void FindSongByNullId(){
        assertThrows(IllegalArgumentException.class, () -> songDao.findById(null));
    }

    @Test
    @Transactional
    void FindSongByNonExistingID() throws Exception {
        assertThrows(Exception.class,()-> songDao.findById(" "));
    }

    @Test
    @Transactional
    void FindSongByName() throws Exception {
        List<Song> songsDb = songDao.findByName(songOne.getSongName());
        List<Song> expectedSongs = new ArrayList<>(List.of(songOne));
        assertEquals(expectedSongs, songsDb);
    }

    @Test
    @Transactional
    void FindSongByNullName() throws Exception {
        assertEquals(new ArrayList<Song>(), songDao.findByName(null));
    }

    @Test
    @Transactional
    void FindSongByEmptyName() throws Exception {
        List<Song> songsDb = songDao.findByName("");
        assertEquals(new ArrayList<>(), songsDb);
    }

    @Test
    @Transactional
    void FindNonExistingSongByName() throws Exception {
        List<Song> songsDb = songDao.findByName("213456");
        assertEquals(new ArrayList<>(), songsDb);
    }

    @Test
    @Transactional
    void UpdateSong() throws Exception {
        Song songFromDb = songDao.findById(songOne.getId());
        songFromDb.setSongName("New song");
        songDao.update(songFromDb);
        Song newSong = songDao.findById(songOne.getId());
        assertEquals("New song", newSong.getSongName());
    }

    @Test
    @Transactional
    void UpdateSongWithNullName() throws Exception {
        Song songInDb = songDao.findById(songOne.getId());
        songInDb.setSongName(null);
        assertThrows(Exception.class,() -> songDao.update(songInDb));
    }

    @Test
    @Transactional
    void UpdateSongWithNullDuration() throws Exception {
        Song songInDb = songDao.findById(songOne.getId());
        songInDb.setDuration(null);
        assertThrows(Exception.class,()-> songDao.update(songInDb));
    }

    @Test
    @Transactional
    void DeleteSong() {
        songDao.delete(songOne);
        assertThrows(Exception.class,()-> songDao.findById(songOne.getId()));
    }

    @Test
    @Transactional
    void DeleteNullSong() {
        assertThrows(Exception.class, () -> songDao.delete(null));
    }

    @Test
    @Transactional
    void DeleteSongNotInDb(){
        assertThrows(Exception.class,  () -> songDao.delete(new Song()));
    }

}
