package cz.fi.muni.pa165.security;

import com.google.inject.internal.util.ImmutableMap;
import cz.fi.muni.pa165.dao.security.UserDao;
import cz.fi.muni.pa165.dto.SystemUserDto;
import cz.fi.muni.pa165.entity.security.User;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataRetrievalFailureException;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;


import java.util.List;
import java.util.Optional;

@Service
public class UserServiceImpl implements UserService{
    private UserDao repository;
    private PasswordService passwordService;
    private TokenService tokens;
    private PasswordEncoder passwordEncoder;

    private UserDao userRepository;

    @Autowired
    public UserServiceImpl(UserDao repository, PasswordService passwordService, TokenService tokens, PasswordEncoder passwordEncoder) {
        this.repository = repository;
        this.passwordService = passwordService;
        this.tokens = tokens;
        this.passwordEncoder = passwordEncoder;
    }


    public UserServiceImpl(UserDao userRepository){
        this.userRepository = userRepository;
    }



    @Override
    public void create(User user) {
        if (user == null) {
            throw new IllegalArgumentException("User is null");
        }
        userRepository.create(user);
    }

    @Override
    public User findById(String id) {
        if (id == null) {
            throw new IllegalArgumentException("Id is null");
        }
        return userRepository.findById(id);
    }

    @Override
    public List<User> findByName(String name) {
        if (name == null) {
            throw new IllegalArgumentException("Name is null");
        }
        return userRepository.findByName(name);
    }

    @Override
    public User update(User user) {
        if (user == null)
            throw new IllegalArgumentException("User is null");

        User userFromDB = userRepository.findById(user.getId());
        if (userFromDB == null)
            throw new DataRetrievalFailureException("User not found in database");

        return userRepository.update(user);
    }

    @Override
    public void delete(User user) {
        if (user == null) {
            throw new IllegalArgumentException("User is null");
        }

        User userFromDB = userRepository.findById(user.getId());
        if (userFromDB == null) {
            throw new DataRetrievalFailureException("User not found in database");
        }
        userRepository.delete(user);
    }

    @Override
    public List<User> findAll() {
        return userRepository.findAll();
    }
    @Override
    public void changePassword(User user, String oldPlainPassword, String newPlainPassword) {
        if (user == null) {
            throw new IllegalArgumentException("User is null");
        }
        if (oldPlainPassword == null) {
            throw new IllegalArgumentException("OldPlainPassword is null");
        }
        if (newPlainPassword == null) {
            throw new IllegalArgumentException("NewPlainPassword is null");
        }
        String newEncodedPassword = passwordService.encodeNewPassword(user, oldPlainPassword, newPlainPassword);
        user.setPassword(newEncodedPassword);
    }

    @Override
    public Optional<String> login(final String username, final String password) {
        if (username == null) {
            throw new IllegalArgumentException("Username is null");
        }
        if (password == null) {
            throw new IllegalArgumentException("Password is null");
        }
        return repository
                .findByLoginName(username)
                .filter(user -> passwordEncoder.matches(password, user.getPassword()))
                .map(user -> tokens.expiring(ImmutableMap.of("username", username)));
    }

    @Override
    public Optional<SystemUserDto> extractUserFromToken(final String token) {
        if (token == null) {
            throw new IllegalArgumentException("Token is null");
        }
        return Optional.of(tokens.verify(token))
                .map(map -> map.get("username"))
                .map(userName -> new SystemUserDto(token, userName));
    }

    @Override
    public User findByLoginName(String loginName) {
        Optional<User> fromDb = repository.findByLoginName(loginName);
        if (fromDb.isEmpty()) throw new IllegalArgumentException(loginName);
        return fromDb.get();
    }

    @Override
    public User getAuthenticatedUser() {
        Authentication authentication = SecurityContextHolder.getContext().getAuthentication();

        if (authentication != null && authentication.getPrincipal() instanceof User) {
            return (User) authentication.getPrincipal();
        }

        throw new IllegalArgumentException("No authenticated user is found.");
    }

    @Override
    public void logout(User user) {
        // Nothing to do
    }


    protected UserDao repository() {
        return repository;
    }
}
