package cz.fi.muni.pa165;

import cz.fi.muni.pa165.ActivityService;
import cz.fi.muni.pa165.ActivityServiceImpl;
import cz.fi.muni.pa165.dao.ActivityDao;
import cz.fi.muni.pa165.entity.Activity;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.junit.jupiter.MockitoSettings;
import org.mockito.quality.Strictness;
import org.springframework.dao.DataRetrievalFailureException;

import java.time.LocalDateTime;
import java.util.List;


@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = Strictness.LENIENT)
public class ActivityServiceTest {

    @Mock
    private ActivityDao dao;

    private ActivityService service;
    private Activity storedActivity;
    private Activity storedActivitySecond;
    private Activity nonExistingActivity;


    private void assignService() {
        service = new ActivityServiceImpl(dao);
    }

    private void mockRepositoryMethods() {
        Mockito.when(dao.findById(storedActivity.getId())).thenReturn(storedActivity);
        Mockito.when(dao.findById(storedActivitySecond.getId())).thenReturn(storedActivitySecond);
        Mockito.when(dao.findAll()).thenReturn(List.of(storedActivity, storedActivitySecond));
        Mockito.when(dao.findByName(storedActivity.getName())).thenReturn(List.of(storedActivity));
    }

    @BeforeEach
    void setUp() {
        assignService();
        storedActivity = new Activity();
        storedActivity.setTime(LocalDateTime.of(2022,3,27,0,0));
        storedActivity.setName("Activity");

        storedActivitySecond = new Activity();
        storedActivitySecond.setTime(LocalDateTime.of(2022,3,27,2,0));
        storedActivitySecond.setName("Activity second");

        nonExistingActivity = new Activity();
        nonExistingActivity.setTime(LocalDateTime.of(2022,3,27,2,0));
        nonExistingActivity.setName("Non Existing Activity");
        mockRepositoryMethods();
    }



    @Test
    void create() {
        service.create(storedActivity);
        Mockito.verify(dao, Mockito.times(1)).create(storedActivity);
    }

    @Test
    void createNull() {
        Assertions.assertThatThrownBy(() -> service.create(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void update() {
        Activity activityToUpdate = service.findById(storedActivity.getId());
        activityToUpdate.setId(storedActivity.getId());
        activityToUpdate.setName("New Name");
        service.update(activityToUpdate);
        Mockito.verify(dao, Mockito.times(1)).update(activityToUpdate);
    }


    @Test
    void updateNull() {
        Assertions.assertThatThrownBy(() -> service.update(null)).isInstanceOf(IllegalArgumentException.class);
    }


    @Test
    void findById() {
        Activity fromRepository = service.findById(storedActivity.getId());
        Assertions.assertThat(fromRepository).isNotNull();
        Assertions.assertThat(fromRepository).usingRecursiveComparison().isEqualTo(storedActivity);
    }

    @Test
    void findByIdNull() {
        Assertions.assertThatThrownBy(() -> service.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findByName(){
        List<Activity> fromRepository = service.findByName(storedActivity.getName());
        Assertions.assertThat(fromRepository).isNotNull();
        Assertions.assertThat(fromRepository).usingRecursiveComparison().isEqualTo(List.of(storedActivity));
    }

    @Test
    void findAll() {
        List<Activity> fromRepository = service.findAll();

        Assertions.assertThat(fromRepository).isNotNull();
        Assertions.assertThat(fromRepository).containsExactlyInAnyOrder(storedActivity, storedActivitySecond);
    }

    @Test
    void delete() {
        service.delete(storedActivity);
        Mockito.verify(dao, Mockito.times(1)).delete(storedActivity);
    }

    @Test
    void deleteNull() {
        Assertions.assertThatThrownBy(() -> service.delete(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void deleteNonExistent() {
        Assertions.assertThatThrownBy(() -> service.delete(nonExistingActivity)).isInstanceOf(DataRetrievalFailureException.class);
    }

}
