package cz.fi.muni.pa165;

import cz.fi.muni.pa165.dao.AlbumDao;
import cz.fi.muni.pa165.entity.Album;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.entity.Song;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;

import org.mockito.junit.jupiter.MockitoSettings;
import org.mockito.quality.Strictness;

import java.time.Duration;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.dao.DataRetrievalFailureException;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.Mockito.*;

/**
 * @author Dávid Želník - 469323
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = Strictness.LENIENT)
public class AlbumServiceTest {

    @Mock
    private AlbumDao albumDao;

    private AlbumService albumService;
    
    private final Album albumOne = new Album();
    private final Album albumTwo = new Album();
    private final Song song = new Song();
    private final Band band = new Band();
    private final Musician musician = new Musician();

    @BeforeEach
    void setUp() throws Exception {
        albumService = new AlbumServiceImpl(albumDao);
        
        song.setSongName("Song name");
        song.setDuration(Duration.ofSeconds(30));

        musician.setName("Musician");
        musician.setType(MusicianType.GUITARIST);

        band.setName("Band");
        band.setType(BandType.ROCK);
        band.setBandMembers(new HashSet<>(List.of(musician)));

        musician.setBand(band);

        albumOne.setName("First");
        albumOne.setBand(band);
        albumOne.setSongs(new HashSet<>(List.of(song)));

        albumTwo.setName("Second");
        albumTwo.setBand(band);
        albumTwo.setSongs(new HashSet<>());

        band.setBandAlbums(new HashSet<>(List.of(albumOne, albumTwo)));

        when(albumDao.findById(albumOne.getId())).thenReturn(albumOne);
        when(albumDao.findById(albumTwo.getId())).thenReturn(albumTwo);

        when(albumDao.findByName(albumOne.getName())).thenReturn(List.of(albumOne));
        when(albumDao.findByName(albumTwo.getName())).thenReturn(List.of(albumTwo));

        when(albumDao.findAll()).thenReturn(List.of(albumOne, albumTwo));
    }

    @Test
    void createAlbum() {
        albumService.create(albumOne);
        verify(albumDao, times(1)).create(albumOne);
    }

    @Test
    void createNullAlbum() {
        assertThatThrownBy(() -> albumService.create(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findAlbumById() {
        Album first = albumService.findById(albumOne.getId());
        Album second = albumService.findById(albumTwo.getId());
        assertThat(first).isNotNull();
        assertThat(first).usingRecursiveComparison().isEqualTo(albumOne);
        assertThat(second).isNotNull();
        assertThat(second).usingRecursiveComparison().isEqualTo(albumTwo);
    }

    @Test
    void findAlbumByNullId(){
        assertThatThrownBy(() -> albumService.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findAlbumByNonExistingID() {
        albumService.findById("notexist");
        verify(albumDao, times(1)).findById("notexist");
    }

    @Test
    void findAlbumByName() {
        albumService.findByName(albumOne.getName());
        verify(albumDao, times(1)).findByName(albumOne.getName());
    }

    @Test
    void findAlbumByNullName() {
        albumService.findByName(null);
        assertThatThrownBy(() -> albumService.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findAlbumByEmptyName() {
        List<Album> albumsDb = albumService.findByName("");
        verify(albumDao, times(1)).findByName("");
        assertThat(albumsDb).isNotNull();
        assertThat(albumsDb).usingRecursiveComparison().isEqualTo(new ArrayList<Album>());
    }

    @Test
    void findNonExistingAlbumByName() {
        List<Album> albumsDb = albumService.findByName("213456");
        verify(albumDao, times(1)).findByName("213456");
        assertThat(albumsDb).isNotNull();
        assertThat(albumsDb).usingRecursiveComparison().isEqualTo(new ArrayList<Album>());
    }

    @Test
    void updateAlbum() {
        Album albumFromDb = albumService.findById(albumOne.getId());
        albumFromDb.setName("New album");
        albumService.update(albumFromDb);
        verify(albumDao, times(1)).update(albumFromDb);
    }

    @Test
    void updateAlbumWithNullName() {
        Album albumInDb = albumService.findById(albumOne.getId());
        albumInDb.setName(null);
        assertThatThrownBy(() -> albumService.update(albumInDb)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void updateAlbumWithNullBandName() {
        Album albumInDb = albumService.findById(albumOne.getId());
        albumInDb.setBand(null);
        assertThatThrownBy(() -> albumService.update(albumInDb)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void updateAlbumSongs() {
        Album albumInDb = albumService.findById(albumOne.getId());
        albumInDb.setSongs(new HashSet<>());
        albumService.update(albumInDb);
        verify(albumDao, times(1)).update(albumInDb);
    }

    @Test
    void deleteAlbum() {
        albumService.delete(albumOne);
        verify(albumDao, times(1)).delete(albumOne);
    }

    @Test
    void deleteNullAlbum() {
        assertThatThrownBy(() -> albumService.delete(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void deleteAlbumNotInDb() {
        assertThatThrownBy(() -> albumService.delete(new Album())).isInstanceOf(DataRetrievalFailureException.class);
    }

}
