package cz.fi.muni.pa165;

import cz.fi.muni.pa165.BandService;
import cz.fi.muni.pa165.BandServiceImpl;
import cz.fi.muni.pa165.dao.BandDao;
import cz.fi.muni.pa165.entity.Activity;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.junit.jupiter.MockitoSettings;
import org.mockito.quality.Strictness;
import org.springframework.dao.DataRetrievalFailureException;

import java.time.LocalDateTime;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = Strictness.LENIENT)
public class BandServiceTest {

    @Mock
    private BandDao dao;

    private BandService service;
    private Band storedBand;
    private Band storedBandSecond;
    private Band nonExistingBand = new Band();
    private final Set<Musician> bandMembers = new HashSet<>();

    private void assignService() {
        service = new BandServiceImpl(dao);
    }

    private void mockRepositoryMethods() {
        Mockito.when(dao.findById(storedBand.getId())).thenReturn(storedBand);
        Mockito.when(dao.findById(storedBandSecond.getId())).thenReturn(storedBandSecond);
        Mockito.when(dao.findAll()).thenReturn(List.of(storedBand, storedBandSecond));
        Mockito.when(dao.findByName(storedBand.getName())).thenReturn(List.of(storedBand));
    }

    @BeforeEach
    void setUp() {
        assignService();

        Musician musicianJane = new Musician();
        musicianJane.setName("Jane Smith");
        musicianJane.setType(MusicianType.SINGER);

        Musician musicianJohn = new Musician();
        musicianJohn.setName("John Doe");
        musicianJohn.setType(MusicianType.GUITARIST);

        bandMembers.add(musicianJane);
        bandMembers.add(musicianJohn);

        storedBand = new Band();
        storedBand.setName("Band");
        storedBand.setType(BandType.ROCK);
        storedBand.setLogo("New Logo");
        storedBand.setBandMembers(bandMembers);

        storedBandSecond = new Band();
        storedBandSecond.setName("Band 2");
        storedBandSecond.setType(BandType.POP);
        storedBandSecond.setLogo("New Logo 2");
        storedBandSecond.setBandMembers(bandMembers);

        mockRepositoryMethods();
    }

    @Test
    void create() {
        service.create(storedBand);
        Mockito.verify(dao, Mockito.times(1)).create(storedBand);
    }

    @Test
    void createNull() {
        Assertions.assertThatThrownBy(() -> service.create(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void update() {
        Band bandToUpdate = service.findById(storedBand.getId());
        bandToUpdate.setId(storedBand.getId());
        bandToUpdate.setName("New Name");
        service.update(bandToUpdate);
        Mockito.verify(dao, Mockito.times(1)).update(bandToUpdate);
    }

    @Test
    void updateNull() {
        Assertions.assertThatThrownBy(() -> service.update(null)).isInstanceOf(IllegalArgumentException.class);
    }


    @Test
    void findById() {
        Band fromRepository = service.findById(storedBand.getId());
        Assertions.assertThat(fromRepository).isNotNull();
        Assertions.assertThat(fromRepository).usingRecursiveComparison().isEqualTo(storedBand);
    }

    @Test
    void findByIdNull() {
        Assertions.assertThatThrownBy(() -> service.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findAll() {
        List<Band> fromRepository = service.findAll();

        Assertions.assertThat(fromRepository).isNotNull();
        Assertions.assertThat(fromRepository).containsExactlyInAnyOrder(storedBand, storedBandSecond);
    }
    @Test
    void findByName(){
        List<Band> fromRepository = service.findByName(storedBand.getName());
        Assertions.assertThat(fromRepository).isNotNull();
        Assertions.assertThat(fromRepository).usingRecursiveComparison().isEqualTo(List.of(storedBand));
    }

    @Test
    void delete() {
        service.delete(storedBand);
        Mockito.verify(dao, Mockito.times(1)).delete(storedBand);
    }

    @Test
    void deleteNull() {
        Assertions.assertThatThrownBy(() -> service.delete(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void deleteNonExistent() {
        Assertions.assertThatThrownBy(() -> service.delete(nonExistingBand)).isInstanceOf(DataRetrievalFailureException.class);
    }
}