package cz.fi.muni.pa165;

import cz.fi.muni.pa165.dao.ManagerDao;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Manager;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.*;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.junit.jupiter.MockitoSettings;
import org.mockito.quality.Strictness;
import org.springframework.dao.DataRetrievalFailureException;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = Strictness.LENIENT)
public class ManagerServiceTest {
    @Mock
    private ManagerDao dao;

    private ManagerService service;
    private Manager storedManager;
    private Manager storedManagerSecond;
    private Band band;
    private Musician musician;

    private final Manager nonExistingManager = new Manager();

    private void assignService() {
        service = new ManagerServiceImpl(dao);
    }

    private void mockRepositoryMethods() {
        Mockito.when(dao.findById(storedManager.getId())).thenReturn(storedManager);
        Mockito.when(dao.findById(storedManagerSecond.getId())).thenReturn(storedManagerSecond);
        Mockito.when(dao.findAll()).thenReturn(List.of(storedManager, storedManagerSecond));
        Mockito.when(dao.findByName(storedManager.getName())).thenReturn(List.of(storedManager));

    }
    @BeforeEach
    void setUp() {
        assignService();

        storedManager = new Manager();
        storedManager.setName("Manager one");

        storedManagerSecond = new Manager();
        storedManagerSecond.setName("Manager two");

        musician = new Musician();
        musician.setName("musician");
        musician.setType(MusicianType.GUITARIST);

        band = new Band();
        band.setName("band");
        band.setType(BandType.ROCK);
        band.setBandMembers(new HashSet<>(List.of(musician)));

        mockRepositoryMethods();
    }
    @AfterAll
    void tearDown() {

    }

    @Test
    void create(){
        service.create(storedManager);
        Mockito.verify(dao, Mockito.times(1)).create(storedManager);
    }
    @Test
    void createNull(){
        Assertions.assertThatThrownBy(() -> service.create(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findById(){
        Manager repo = service.findById(storedManager.getId());
        Assertions.assertThat(repo).isNotNull();
        Assertions.assertThat(repo).usingRecursiveComparison().isEqualTo(storedManager);
    }

    @Test
    void findByNullId(){
        Assertions.assertThatThrownBy(() -> service.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findAll(){
        List<Manager> repo = service.findAll();

        Assertions.assertThat(repo).isNotNull();
        Assertions.assertThat(repo).containsExactlyInAnyOrder(storedManager, storedManagerSecond);
    }

    @Test
    void findByName(){
        List<Manager> repo = service.findByName(storedManager.getName());

        Assertions.assertThat(repo).isNotNull();
        Assertions.assertThat(repo).usingRecursiveComparison().isEqualTo(List.of(storedManager));
    }

    @Test
    void delete() {
        service.delete(storedManager);
        Mockito.verify(dao, Mockito.times(1)).delete(storedManager);
    }

    @Test
    void deleteNull() {
        Assertions.assertThatThrownBy(() -> service.delete(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void deleteNonExistent() {
        Assertions.assertThatThrownBy(() -> service.delete(nonExistingManager)).isInstanceOf(DataRetrievalFailureException.class);
    }

    @Test
    void addManagedBandNullBand(){
        Assertions.assertThatThrownBy(() -> service.addManagedBand(null, storedManager)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void addManagedBandNullManager(){
        Assertions.assertThatThrownBy(() -> service.addManagedBand(band, null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void addManagedBand(){
        Set<Band> bands = service.addManagedBand(band, storedManager);
        Manager manager = service.findById(storedManager.getId());
        Assertions.assertThat(manager.getManagedBands()).isEqualTo(bands);
    }
}
