package cz.fi.muni.pa165;

import cz.fi.muni.pa165.dao.MusicianDao;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.entity.Offer;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.junit.jupiter.MockitoSettings;
import org.mockito.quality.Strictness;
import org.springframework.dao.DataRetrievalFailureException;

import java.util.List;


@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = Strictness.LENIENT)
public class MusicianServiceTest {
    @Mock
    private MusicianDao dao;

    private MusicianService service;
    private Musician storedMusician;
    private Musician storedMusicianTwo;
    private Musician nonExistingMusician = new Musician();
    private Offer offer;
    private Band band;

    private void assignService() {
        service = new MusicianServiceImpl(dao);
    }

    private void mockRepositoryMethods() {
        Mockito.when(dao.findById(storedMusician.getId())).thenReturn(storedMusician);
        Mockito.when(dao.findById(storedMusicianTwo.getId())).thenReturn(storedMusicianTwo);
        Mockito.when(dao.findAll()).thenReturn(List.of(storedMusician, storedMusicianTwo));
        Mockito.when(dao.findByName(storedMusician.getName())).thenReturn(List.of(storedMusician));
        Mockito.when(dao.findAvailableMusicians()).thenReturn(List.of(storedMusician, storedMusicianTwo));
    }

    @BeforeEach
    void setUp() {
        assignService();
        storedMusician = new Musician();
        storedMusician.setName("John");
        storedMusician.setType(MusicianType.DRUMMER);

        storedMusicianTwo = new Musician();
        storedMusicianTwo.setName("Paul");
        storedMusicianTwo.setType(MusicianType.SINGER);

        band = new Band();
        band.setName("New band");
        band.setType(BandType.ROCK);

        offer = new Offer();
        offer.setName("New offer");
        offer.setBand(band);

        mockRepositoryMethods();
    }

    @Test
    void create() {
        service.create(storedMusician);
        Mockito.verify(dao, Mockito.times(1)).create(storedMusician);
    }

    @Test
    void createNull() {
        Assertions.assertThatThrownBy(() -> service.create(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void update() {
        Musician musicianToUpdate = service.findById(storedMusician.getId());
        musicianToUpdate.setId(storedMusician.getId());
        musicianToUpdate.setName("New Name");
        service.update(musicianToUpdate);
        Mockito.verify(dao, Mockito.times(1)).update(musicianToUpdate);
    }

    @Test
    void updateNull() {
        Assertions.assertThatThrownBy(() -> service.update(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findById() {
        Musician musicianFromRepo = service.findById(storedMusician.getId());
        Assertions.assertThat(musicianFromRepo).isNotNull();
        Assertions.assertThat(musicianFromRepo).usingRecursiveComparison().isEqualTo(storedMusician);
    }

    @Test
    void findByIdNull() {
        Assertions.assertThatThrownBy(() -> service.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findAll() {
        List<Musician> musiciansFromRepo = service.findAll();
        Assertions.assertThat(musiciansFromRepo).isNotNull();
        Assertions.assertThat(musiciansFromRepo).containsExactlyInAnyOrder(storedMusician, storedMusicianTwo);
    }

    @Test
    void findByName(){
        List<Musician> musiciansFromRepo = service.findByName(storedMusician.getName());
        Assertions.assertThat(musiciansFromRepo).isNotNull();
        Assertions.assertThat(musiciansFromRepo).usingRecursiveComparison().isEqualTo(List.of(storedMusician));
    }

    @Test
    void delete() {
        service.delete(storedMusician);
        Mockito.verify(dao, Mockito.times(1)).delete(storedMusician);
    }

    @Test
    void deleteNull() {
        Assertions.assertThatThrownBy(() -> service.delete(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void deleteNonExistent() {
        Assertions.assertThatThrownBy(() -> service.delete(nonExistingMusician))
                .isInstanceOf(DataRetrievalFailureException.class);
    }

    @Test
    void acceptOffer() {
        service.acceptOffer(offer, storedMusician);
        Band musiciansBand = storedMusician.getBand();
        Assertions.assertThat(band).usingRecursiveComparison().isEqualTo(musiciansBand);
    }

    @Test
    void acceptNullOffer() {
        Assertions.assertThatThrownBy(() -> service.acceptOffer(null, storedMusician))
                .isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void acceptOfferNullMusician() {
        Assertions.assertThatThrownBy(() -> service.acceptOffer(offer, null))
                .isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findAvailableMusicians(){
        Assertions.assertThat(service.findAvailableMusicians())
                .isEqualTo(List.of(storedMusician, storedMusicianTwo));
    }
}
