package cz.fi.muni.pa165;

import cz.fi.muni.pa165.dao.OfferDao;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Offer;
import cz.fi.muni.pa165.enums.BandType;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.junit.jupiter.MockitoSettings;
import org.mockito.quality.Strictness;
import org.springframework.dao.DataRetrievalFailureException;

import java.util.List;


@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = Strictness.LENIENT)
public class OfferServiceTest {

    @Mock
    private OfferDao dao;

    private OfferService service;
    private Offer storedOffer;
    private Offer storedOfferSecond;
    private Offer nonExistingOffer;


    private void assignService() {
        service = new OfferServiceImpl(dao);
    }

    private void mockRepositoryMethods() {
        Mockito.when(dao.findById(storedOffer.getId())).thenReturn(storedOffer);
        Mockito.when(dao.findById(storedOfferSecond.getId())).thenReturn(storedOfferSecond);
        Mockito.when(dao.findAll()).thenReturn(List.of(storedOffer, storedOfferSecond));
        Mockito.when(dao.findByName(storedOffer.getName())).thenReturn(List.of(storedOffer));
    }

    @BeforeEach
    void setUp() {
        assignService();
        Band band = new Band();
        band.setName("New band");
        band.setType(BandType.ROCK);

        storedOffer = new Offer();
        storedOffer.setName("New offer");
        storedOffer.setBand(band);

        storedOfferSecond = new Offer();
        storedOfferSecond.setName("New offer 2");

        nonExistingOffer = new Offer();

        mockRepositoryMethods();
    }

    @Test
    void create() {
        service.create(storedOffer);
        Mockito.verify(dao, Mockito.times(1)).create(storedOffer);
    }

    @Test
    void createNull() {
        Assertions.assertThatThrownBy(() -> service.create(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void update() {
        Offer offerToUpdate = service.findById(storedOffer.getId());
        offerToUpdate.setId(storedOffer.getId());
        offerToUpdate.setName("New Name");
        service.update(offerToUpdate);
        Mockito.verify(dao, Mockito.times(1)).update(offerToUpdate);
    }

    @Test
    void updateNull() {
        Assertions.assertThatThrownBy(() -> service.update(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findById() {
        Offer fromRepo = service.findById(storedOffer.getId());
        Assertions.assertThat(fromRepo).isNotNull();
        Assertions.assertThat(fromRepo).usingRecursiveComparison().isEqualTo(storedOffer);
    }

    @Test
    void findByIdNull() {
        Assertions.assertThatThrownBy(() -> service.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findByName(){
        List<Offer> fromRepo = service.findByName(storedOffer.getName());
        Assertions.assertThat(fromRepo).isNotNull();
        Assertions.assertThat(fromRepo).usingRecursiveComparison().isEqualTo(List.of(storedOffer));
    }

    @Test
    void findAll() {
        List<Offer> fromRepo = service.findAll();
        Assertions.assertThat(fromRepo).isNotNull();
        Assertions.assertThat(fromRepo).containsExactlyInAnyOrder(storedOffer, storedOfferSecond);
    }

    @Test
    void delete() {
        service.delete(storedOffer);
        Mockito.verify(dao, Mockito.times(1)).delete(storedOffer);
    }

    @Test
    void deleteNull() {
        Assertions.assertThatThrownBy(() -> service.delete(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void deleteNonExistent() {
        Assertions.assertThatThrownBy(() -> service.delete(nonExistingOffer))
                .isInstanceOf(DataRetrievalFailureException.class);
    }
}
