package cz.fi.muni.pa165;

import cz.fi.muni.pa165.dao.SongDao;
import cz.fi.muni.pa165.entity.Song;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;

import org.mockito.junit.jupiter.MockitoSettings;
import org.mockito.quality.Strictness;

import java.time.Duration;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.dao.DataRetrievalFailureException;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

import static org.mockito.Mockito.*;

/**
 * @author Dávid Želník - 469323
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = Strictness.LENIENT)
public class SongServiceTest {

    @Mock
    private SongDao songDao;

    private SongService songService;

    private final Song songOne = new Song();
    private final Song songTwo = new Song();
    private final Band band = new Band();
    private final Musician musician = new Musician();

    @BeforeEach
    void setUp(){
        songService = new SongServiceImpl(songDao);
        musician.setName("Musician");
        musician.setType(MusicianType.GUITARIST);

        songOne.setSongName("Song One");
        songOne.setDuration(Duration.ofSeconds(30));
        songTwo.setSongName("Song Two");
        songTwo.setDuration(Duration.ofSeconds(30));

        band.setName("Band");
        band.setType(BandType.ROCK);
        band.setBandMembers(new HashSet<>(List.of(musician)));

        when(songDao.findById(songOne.getId())).thenReturn(songOne);
        when(songDao.findById(songTwo.getId())).thenReturn(songTwo);

        when(songDao.findByName(songOne.getSongName())).thenReturn(List.of(songOne));
        when(songDao.findByName(songTwo.getSongName())).thenReturn(List.of(songTwo));

        when(songDao.findAll()).thenReturn(List.of(songOne, songTwo));
    }

    @Test
    void createSong() {
        songDao.create(songOne);
        verify(songDao, times(1)).create(songOne);
    }

    @Test
    void createNullSong() {
        assertThatThrownBy(() -> songService.create(null)).isInstanceOf(IllegalArgumentException.class);
    }
    
    @Test
    void findSongById() {
        Song first = songDao.findById(songOne.getId());
        Song second = songDao.findById(songTwo.getId());
        assertThat(first).isNotNull();
        assertThat(first).usingRecursiveComparison().isEqualTo(songOne);
        assertThat(second).isNotNull();
        assertThat(second).usingRecursiveComparison().isEqualTo(songTwo);
    }

    @Test
    void findSongByNullId(){
        assertThatThrownBy(() -> songService.findById(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findSongByNonExistingID() {
        songService.findById("notexist");
        verify(songDao, times(1)).findById("notexist");
    }

    @Test
    void findSongByName() {
        songService.findByName(songOne.getSongName());
        verify(songDao, times(1)).findByName(songOne.getSongName());
    }

    @Test
    void findSongByNullName() {
        assertThatThrownBy(() -> songService.findByName(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void findSongByEmptyName() {
        List<Song> songsDb = songDao.findByName("");
        assertThat(songsDb).isNotNull();
        assertThat(songsDb).usingRecursiveComparison().isEqualTo(new ArrayList<Song>());
    }

    @Test
    void findNonExistingSongByName() {
        List<Song> songsDb = songDao.findByName("213456");
        assertThat(songsDb).usingRecursiveComparison().isEqualTo(new ArrayList<Song>());
    }

    @Test
    void updateSong() {
        songOne.setSongName("New song");
        songService.update(songOne);
        verify(songDao, times(1)).update(songOne);
    }


    @Test
    void deleteSong() {
        songService.delete(songOne);
        verify(songDao, times(1)).delete(songOne);
    }

    @Test
    void deleteNullSong() {
        assertThatThrownBy(() -> songService.delete(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void deleteSongNotInDb(){
        assertThatThrownBy(() -> songService.delete(new Song())).isInstanceOf(DataRetrievalFailureException.class);
    }

}
